/******************************************************************************
 * JBoss, a division of Red Hat                                               *
 * Copyright 2006, Red Hat Middleware, LLC, and individual                    *
 * contributors as indicated by the @authors tag. See the                     *
 * copyright.txt in the distribution for a full listing of                    *
 * individual contributors.                                                   *
 *                                                                            *
 * This is free software; you can redistribute it and/or modify it            *
 * under the terms of the GNU Lesser General Public License as                *
 * published by the Free Software Foundation; either version 2.1 of           *
 * the License, or (at your option) any later version.                        *
 *                                                                            *
 * This software is distributed in the hope that it will be useful,           *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU           *
 * Lesser General Public License for more details.                            *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public           *
 * License along with this software; if not, write to the Free                *
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA         *
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.                   *
 ******************************************************************************/
package org.jboss.portletbridge.context;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.security.Principal;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.faces.FacesException;
import javax.faces.application.ViewHandler;
import javax.faces.component.UIViewRoot;
import javax.faces.context.ExternalContext;
import javax.faces.context.FacesContext;
import javax.faces.context.FacesContextFactory;
import javax.faces.render.ResponseStateManager;
import javax.portlet.BaseURL;
import javax.portlet.ClientDataRequest;
import javax.portlet.MimeResponse;
import javax.portlet.PortletContext;
import javax.portlet.PortletException;
import javax.portlet.PortletRequest;
import javax.portlet.PortletRequestDispatcher;
import javax.portlet.PortletResponse;
import javax.portlet.PortletSession;
import javax.portlet.faces.Bridge;
import javax.portlet.faces.BridgeDefaultViewNotSpecifiedException;
import javax.portlet.faces.BridgeInvalidViewPathException;

import org.jboss.portletbridge.BridgeRequestScope;
import org.jboss.portletbridge.util.BridgeLogger;

/**
 * Version of the {@link ExternalContext} for a Portlet request.
 * {@link FacesContextFactory} will create instance of this class for a portal
 * <code>action</code> phase.
 * 
 * @author asmirnov
 * 
 */
public abstract class PortletExternalContextImpl extends
		AbstractExternalContext {

	public static final String SERVLET_PATH_ATTRIBUTE = "javax.servlet.include.servlet_path";
	public static final String PATH_INFO_ATTRIBUTE = "javax.servlet.include.path_info";
	public static final String ACTION_URL_DO_NOTHITG = "/JBossPortletBridge/actionUrl/do/nothing";
	public static final String RESOURCE_URL_DO_NOTHITG = "/JBossPortletBridge/resourceUrl/do/nothing";
	public static final String WSRP_REWRITE = "wsrp_rewrite?";
	/**
	 * Request parameter to store current View Id.
	 */
	public static final String VIEW_ID_PARAMETER = "org.jboss.portletbridge.VIEWID";
	protected static final String[] EMPTY_STRING_ARRAY = new String[0];
	public static final String PORTLET_CONFIG_ATTRIBUTE = "org.jboss.portletbridge.CONFIG";
	public static final String INITIAL_REQUEST_ATTRIBUTES_NAMES = "org.jboss.portletbridge.REQUEST_PARAMETERS";
	public static final Object RENDER_POLICY_ATTRIBUTE = "org.jboss.portletbridge.RENDER_POLICY";
	public static final String PORTAL_USER_PRINCIPAL = "org.jboss.portletbridge.USER_PRINCIPAL";
	public static final String CONVERSATION_ID_PARAMETER = "conversationId";
	protected static final Logger LOGGER = BridgeLogger.CONTEXT.getLogger();

	private String servletPath = null;
	private String servletMappingSuffix;
	private String defaultJsfSuffix;
	private String servletMappingPrefix;
	private String viewId;
	private String pathInfo = null;
	private final Map<String,Map<String,String[]>> encodedActionUrlParameters = new HashMap<String, Map<String, String[]>>();
	private final Map<String, String[]> extraRequestParameters = new HashMap<String, String[]>();
	protected final PortletBridgeContext portletBridgeContext;

	enum Scheme {
		action, render, resource
	}

	public PortletExternalContextImpl(PortletContext context,
			PortletRequest request, PortletResponse response) {
		super(context, request, response);
		portletBridgeContext = (PortletBridgeContext) request
				.getAttribute(PortletBridgeContext.REQUEST_PARAMETER_NAME);
		if (null == portletBridgeContext) {
			throw new FacesException("No portlet bridge context");
		}
		defaultJsfSuffix = context.getInitParameter(
				ViewHandler.DEFAULT_SUFFIX_PARAM_NAME);
		if (null == defaultJsfSuffix) {
			defaultJsfSuffix = ViewHandler.DEFAULT_SUFFIX;
		}
		// Calculate FacesServlet prefix.
		calculateViewId();
		// Set render kit parameter as required by JSR pp 6.9
		String defaultRenderKitId = portletBridgeContext.getBridgeConfig().getDefaultRenderKitId();
		if(null != defaultRenderKitId && null == request.getParameter(ResponseStateManager.RENDER_KIT_ID_PARAM)){
			extraRequestParameters.put(ResponseStateManager.RENDER_KIT_ID_PARAM, new String[]{defaultRenderKitId});
		}
	}

	public void setResponseCharacterEncoding(String encoding) {
		// Do nothing
	}

	public String getResponseCharacterEncoding() {
		throw new IllegalStateException(
				"responseCharacterEncoding undefined for request");
	}

	@Override
    public String getResponseContentType()
    {
        if (!(getResponse() instanceof MimeResponse))
        {
          throw new IllegalStateException("PortletExternalContextImpl.getResponseContentType(): Response must be a MimeResponse");
        }

    return ((MimeResponse) getResponse()).getContentType();
    }

	@Override
    public String getRequestContentType()
    {
    if (getRequestMap().get(Bridge.PORTLET_LIFECYCLE_PHASE) != null && getRequest() instanceof ClientDataRequest)
    {
      return ((ClientDataRequest) getRequest()).getContentType();
    }
    else
    {
      // Other PHASEs: return null as per spec
      return null;
    }
    }

	@Override
    public String getRequestCharacterEncoding()
    {
    if (getRequestMap().get(Bridge.PORTLET_LIFECYCLE_PHASE) != null && getRequest() instanceof ClientDataRequest)
    {
      return ((ClientDataRequest) getRequest()).getCharacterEncoding();
    }
    else
    {
      // other phases -- return null as per spec
      return null;
    }
    }

    @Override
    public void setRequestCharacterEncoding(String encoding)
    throws UnsupportedEncodingException, IllegalStateException
    {
    if (getRequestMap().get(Bridge.PORTLET_LIFECYCLE_PHASE) != null && getRequest() instanceof ClientDataRequest)
    {
      try
      {
        ((ClientDataRequest) getRequest()).setCharacterEncoding(encoding);
      } catch (IllegalStateException e)
      {
        // Swallow this exception and proceed as if a noop.
      }
    }

    // Though can only do this operation in a PortletAction Faces impls have
    // been known to cache the encoding in memory and set in situations when
    // the request doesn't specify -- turns out this code can get hit during
    // the bridge render phase -- so if we threw an exception Faces would break.
    // Just do a noop instead as render isn't supposed to receive parameters
    }

	public PortletContext getContext() {
		return (PortletContext) super.getContext();
	}

	public PortletRequest getRequest() {
		return (PortletRequest) super.getRequest();
	}

	public PortletResponse getResponse() {
		return (PortletResponse) super.getResponse();
	}

	public String getInitParameter(String name) {
		return getContext().getInitParameter(name);
	}

	protected String getNamespace() {
		return getResponse().getNamespace();
	}

	public URL getResource(String path) throws MalformedURLException {
		return getContext().getResource(path);
	}

	public InputStream getResourceAsStream(String path) {
		return getContext().getResourceAsStream(path);
	}

	public Set<String> getResourcePaths(String path) {
		return getContext().getResourcePaths(path);
	}

	protected Object getContextAttribute(String name) {
		return getContext().getAttribute(name);
	}

	protected Enumeration<String> getContextAttributeNames() {
		return getContext().getAttributeNames();
	}

	protected Enumeration<String> getInitParametersNames() {
		return getContext().getInitParameterNames();
	}

	protected Object getRequestAttribute(String name) {
		if (PATH_INFO_ATTRIBUTE.equals(name)) {
			return getRequestPathInfo();
		} else if (SERVLET_PATH_ATTRIBUTE.equals(name)) {
			return getRequestServletPath();
		} else {
			return getRequest().getAttribute(name);
		}
	}

	protected Enumeration<String> getRequestAttributeNames() {
		return getRequest().getAttributeNames();
	}

	protected Map<String, String[]> getNativeRequestParameterValuesMap() {
		return getRequest().getParameterMap();
	}
	
	@Override
	protected Map<String, String[]> getExtraRequestParameterValuesMap() {
		return extraRequestParameters;
	}


	protected Object getSessionAttribute(String name) {
		return getSessionAttribute(name, getScopeForName(name));
	}

	protected int getScopeForName(String name) {
		return this.portletBridgeContext.getBridgeConfig().getStrategy()
				.getPortletSessionScopeForName(name);
	}

	protected Object getSessionAttribute(String name, int scope) {
		return getRequest().getPortletSession(true).getAttribute(name, scope);
	}

	protected Enumeration<String> getSessionAttributeNames() {
		class AttributeEnumeration implements Enumeration<String> {
			int scope;
			Enumeration<String> attributes;

			public AttributeEnumeration() {
				scope = PortletSession.PORTLET_SCOPE;
				attributes = getSessionAttributeNames(scope);
				if (!attributes.hasMoreElements()) {
					scope = PortletSession.APPLICATION_SCOPE;
					attributes = getSessionAttributeNames(scope);
				}
			}

			public boolean hasMoreElements() {
				return attributes.hasMoreElements();
			}

			public String nextElement() {
				final String result = attributes.nextElement();

				if (!attributes.hasMoreElements()
						&& scope == PortletSession.PORTLET_SCOPE) {
					scope = PortletSession.APPLICATION_SCOPE;
					attributes = getSessionAttributeNames(scope);
				}

				return result;
			}
		}

		return new AttributeEnumeration();

		// return getSessionAttributeNames(PortletSession.PORTLET_SCOPE);
	}

	protected Enumeration<String> getSessionAttributeNames(int scope) {
		return getRequest().getPortletSession(true).getAttributeNames(scope);
	}

	protected void removeContextAttribute(String name) {
		getContext().removeAttribute(name);
	}

	protected void removeRequestAttribute(String name) {
		getRequest().removeAttribute(name);
	}

	protected void removeSessionAttribute(String name) {
		removeSessionAttribute(name, getScopeForName(name));
	}

	protected void removeSessionAttribute(String name, int scope) {
		getRequest().getPortletSession(true).removeAttribute(name, scope);
	}

	protected void setContextAttribute(String name, Object value) {
		getContext().setAttribute(name, value);
	}

	protected void setRequestAttribute(String name, Object value) {
		getRequest().setAttribute(name, value);
	}

	protected void setSessionAttribute(String name, Object value) {
		setSessionAttribute(name, value, getScopeForName(name));
	}

	protected void setSessionAttribute(String name, Object value, int scope) {
		getRequest().getPortletSession(true).setAttribute(name, value, scope);
	}

	// private static final Pattern absoluteUrl = Pattern.compile("");
	// private static final Pattern directLink = Pattern.compile("");
	// private static final Pattern viewIdParam = Pattern.compile("");

	/**
	 * @param url
	 * @return
	 */
	protected String encodeURL(String url) {
		try {
			return getResponse().encodeURL(url);
		} catch (IllegalArgumentException e) {
			LOGGER.log(Level.WARNING, "Invalid url for encode:" + url, e);
			throw e;
		}
	}

	protected String encodeURL(PortalActionURL portalUrl, boolean escape) {
		String encodeURL = encodeURL(portalUrl.toString());
		return escapeUrl(escape, encodeURL);
	}

	public String getAuthType() {
		return getRequest().getAuthType();
	}

	public String getRemoteUser() {
		String user = getRequest().getRemoteUser();
		if (user == null) {
			Principal userPrincipal = getUserPrincipal();
			if (null != userPrincipal) {
				user = userPrincipal.getName();

			}
		}
		return user;
	}

	public String getRequestContextPath() {
		return getRequest().getContextPath();
	}

	public Locale getRequestLocale() {
		return getRequest().getLocale();
	}

	public Iterator<Locale> getRequestLocales() {
		return new EnumerationIterator<Locale>(getRequest().getLocales());
	}

	public String getRequestPathInfo() {
		return pathInfo;
	}

	public String getRequestServletPath() {
		return servletPath;
	}

	public Object getSession(boolean create) {
		return getRequest().getPortletSession(create);
	}

	public Principal getUserPrincipal() {
		return getRequest().getUserPrincipal();
	}

	public boolean isUserInRole(String role) {
		return getRequest().isUserInRole(role);
	}

	public void log(String message) {
		getContext().log(message);
	}

	public void log(String message, Throwable exception) {
		getContext().log(message, exception);
	}

	protected void calculateViewId() {
		String viewIdPath = (String) getRequest().getAttribute(Bridge.VIEW_ID);

		String portletModeName = getRequest().getPortletMode().toString();
		List<String> servletMappings = portletBridgeContext.getBridgeConfig()
		.getFacesServletMappings();
		if (null == viewIdPath) {
			String viewPath = (String) getRequest().getAttribute(
					Bridge.VIEW_PATH);
			if (null != viewPath) {
				viewPath = processViewIdParameters(viewPath);
				// Find mapping for view path.
				for (String mapping : servletMappings) {
					if(mapping.startsWith("*.")){
						String 	suffix = mapping.substring(mapping.indexOf('.'));
						if(viewPath.endsWith(suffix)){
							int i = viewPath.lastIndexOf(suffix);
							viewId = viewPath.substring(0, i) + getDefaultJsfSuffix();
							break;
						}
					} else if(mapping.endsWith("/*")){
						String prefix = mapping.substring(0, mapping.length()-1);
						if(viewPath.startsWith(prefix)){
							viewId = viewPath.substring(mapping.length()-2);
							break;
						}
					}
				}
				if(null == viewId){
					throw new BridgeInvalidViewPathException(viewPath);
				}
			} else {
				// Get stored ViewId from window state
				BridgeRequestScope windowState = portletBridgeContext
						.getRequestScope();
				if (portletBridgeContext.isRestoreViewId()) {
					if (null != windowState && null != windowState.getViewId()) {
						viewId = windowState.getViewId();
					} else {
						// Try to get viewId from stored session attribute
						PortletSession portletSession = getRequest()
								.getPortletSession(false);
						if (null != portletSession) {
							viewIdPath = (String) portletSession
									.getAttribute(Bridge.VIEWID_HISTORY + "."
											+ portletModeName);
						}
						if (null == viewIdPath) {
							viewIdPath = calculateDefaultViewId(portletModeName);
						}
						if (null == viewIdPath) {
							throw new BridgeDefaultViewNotSpecifiedException();
						}
						viewId = processViewIdParameters(viewIdPath);
					}
				}
			}
		} else {
			viewId = processViewIdParameters(viewIdPath);
		}

        //remove viewId so eventrequest redirects can be handled properly
        getRequest().removeAttribute(Bridge.VIEW_ID);
		// detect mapping and convert viewId into appropriate url
        calculateServletPath(viewId, servletMappings);
	}

	private String calculateDefaultViewId(String modeName) {
		String defaultViewId = portletBridgeContext.getBridgeConfig().getDefaultViewIdMap()
				.get(modeName);
		return defaultViewId;
	}

	private String processViewIdParameters(String defaultViewId) {
		String viewId;
		try {
			PortalActionURL viewIdUrl = new PortalActionURL(
					defaultViewId);
			viewId = viewIdUrl.getPath();
			getExtraRequestParameterValuesMap().putAll(viewIdUrl.getParameters());
		} catch (MalformedURLException e) {
			viewId = defaultViewId;
		}
		return viewId;
	}

	protected void calculateServletPath(String viewId,
			List<String> servletMappings) {
		if (null != servletMappings && servletMappings.size() > 0) {
			String mapping = servletMappings.get(0);
			if (mapping.startsWith("*")) {
				// Suffix Mapping
				servletMappingSuffix = mapping.substring(mapping.indexOf('.'));
				int i = viewId.lastIndexOf(defaultJsfSuffix);
				if (i < 0) {
					i = viewId.lastIndexOf('.');
				}
				if (i >= 0) {
					viewId = viewId.substring(0, i) + servletMappingSuffix;
				}
				servletPath = viewId;
				getRequest().setAttribute(SERVLET_PATH_ATTRIBUTE, viewId);
				getRequest().setAttribute("com.sun.faces.INVOCATION_PATH",
						servletMappingSuffix);
				pathInfo = null;
			} else if (mapping.endsWith("*")) {
				mapping = mapping.substring(0, mapping.length() - 1);
				if (mapping.endsWith("/")) {
					mapping = mapping.substring(0, mapping.length() - 1);
				}
				servletMappingPrefix = servletPath = mapping;
				getRequest().setAttribute("com.sun.faces.INVOCATION_PATH",
						mapping);
				pathInfo = viewId;
			} else {
				servletPath = null;
				pathInfo = viewId;
			}
		} else {
			servletPath = null;
			pathInfo = viewId;
		}
	}

	/**
	 * @return the servletMappingSuffix
	 */
	public String getServletMappingSuffix() {
		return servletMappingSuffix;
	}

	/**
	 * @return the defaultJsfSuffix
	 */
	public String getDefaultJsfSuffix() {
		return defaultJsfSuffix;
	}

	/**
	 * @return the defaultJsfPrefix
	 */
	public String getServletMappingPrefix() {
		return servletMappingPrefix;
	}

	protected String getViewIdFromPath(String pathInContext) {
		String viewId;
		if (null != getServletMappingPrefix()
				&& pathInContext.startsWith(getServletMappingPrefix())) {
			viewId = pathInContext
					.substring(getServletMappingPrefix().length());
		} else if (null != getServletMappingSuffix()
				&& pathInContext.endsWith(getServletMappingSuffix())) {
			int i = pathInContext.lastIndexOf(getServletMappingSuffix());
			viewId = pathInContext.substring(0, i) + getDefaultJsfSuffix();
		} else {
//			throw new BridgeInvalidViewPathException(pathInContext);
			viewId = pathInContext;
		}
		return viewId;
	}

	public void dispatch(String path) throws IOException {
		if (null == path) {
			throw new NullPointerException("Path to new view is null");
		}
		if (LOGGER.isLoggable(Level.FINE)) {
			LOGGER.fine("Dispatch request to the " + path);
		}
		PortletRequestDispatcher dispatcher = getContext()
				.getRequestDispatcher(path);
		if (null == dispatcher) {
			throw new IllegalStateException(
					"Dispatcher for render request is not created");
		}
		// Bridge has had to set this attribute so Faces RI will skip servlet
		// dependent
		// code when mapping from request paths to viewIds -- however we need to
		// remove it
		// as it screws up the dispatch
		// Object oldPath = getRequestMap().remove(SERVLET_PATH_ATTRIBUTE);
		try {
			if(this.portletBridgeContext.isForwarded()){
				dispatcher.include(getRequest(), getResponse());
			} else {
				dispatcher.forward(getRequest(), getResponse());
				this.portletBridgeContext.setForwarded(true);
			}
		} catch (PortletException e) {
			throw new FacesException(e);
		} finally {
			// if(null !=oldPath){
			// getRequestMap().put(SERVLET_PATH_ATTRIBUTE, oldPath);
			// }
		}
	}

	public void redirect(String url) throws IOException {
		if (null == url || url.length() == 0) {
			throw new NullPointerException("Path to redirect is null");
		}
		if(encodedActionUrlParameters.containsKey(url)){
			// Already encoded URL, use saved parameters
			internalRedirect(url);
		} else if (url.startsWith("#")) {
			externalRedirect(url);
		} else {
				PortalActionURL actionUrl = new PortalActionURL(url);
				if(Boolean.parseBoolean(actionUrl.getParameter(Bridge.DIRECT_LINK))||!isInContext(actionUrl)){
					externalRedirect(url);
				} else {
					String encodedActionURL = encodeActionURL(url);
					internalRedirect(encodedActionURL);
				}
		}
        FacesContext.getCurrentInstance().responseComplete();
	}

	protected abstract void externalRedirect(String url) throws IOException;

	/**
	 * @param actionURL
	 * @throws IOException 
	 */
	protected void internalRedirect(String url) throws IOException {
		// Detect ViewId from URL and create new view for them.
		Map<String, String[]> requestParameters = encodedActionUrlParameters.get(url);
		if (requestParameters.containsKey(Bridge.FACES_VIEW_ID_PARAMETER)) {
			// Save new viewId to restore after redirect.
			String viewId = requestParameters.get(Bridge.FACES_VIEW_ID_PARAMETER)[0];
			portletBridgeContext.setRedirectViewId(viewId);
			portletBridgeContext
						.setRedirectRequestParameters(requestParameters);
	
		} else {
			internalRedirectToNonFacesView(url, requestParameters);
		}
	}

	protected abstract void internalRedirectToNonFacesView(
			String url, Map<String, String[]> requestParameters) throws IOException;

	public String encodeActionURL(String url) {
		if (null == url) {
			throw new NullPointerException();
		}
		if (LOGGER.isLoggable(Level.FINE)) {
			LOGGER.fine("Encode action URL:" + url);
		}
		String actionUrl;
		Map<String, String[]> actionParameters;
		if (url.startsWith("#")) {
			/*
			 * If the inputURL starts with the # character[6.4] return the
			 * inputURL unchanged.
			 */
			actionUrl = url;
			actionParameters = Collections.emptyMap();
		} else if (url.startsWith(PortletExternalContextImpl.WSRP_REWRITE)) {
			// WSRP encoded URL, may need xml escape
			actionUrl = url;
			actionParameters = Collections.emptyMap();
		} else {
			try {
				boolean escapedUrl = isStrictEscaped(url);
				PortalActionURL portalUrl = new PortalActionURL(escapedUrl?unescapeUrl(url):url);
				if (portalUrl.isAbsolute()) {
					/*
					 * If the inputURL starts with the URI scheme "portlet:" the
					 * target is the portlet itself. Interpret this inputURL as
					 * follows: The scheme "portlet:" indicates that the target
					 * of this action is the portlet itself. Though generally
					 * used to generate links to nonFaces views in this portlet
					 * it can also be used to generate action or render links to
					 * a Faces view (including the current view).
					 */
					if ("portlet:".equals(portalUrl.getProtocol())) {
						/*
						 * * The scheme "portlet:" indicates that the target of
						 * this action is the portlet itself. Though generally
						 * used to generate links to nonFaces views in this
						 * portlet it can also be used to generate action or
						 * render links to a Faces view (including the current
						 * view). The scheme is followed by either the keyword
						 * action, render or resource. render indicates a
						 * portlet renderURL should be encoded[6.8]. action
						 * indicates a portlet actionURL should be encoded[6.9].
						 * resource indicates a portlet resourceURL should be
						 * encoded[6.102]. Following this url type indicator is
						 * an optional query string. Parameter value pairs in
						 * the query string are the parameters that are to be
						 * encoded into the portletURL.
						 */
						try {
							if (LOGGER.isLoggable(Level.FINE)) {
								LOGGER.fine("Process URL for 'portlet:' protocol ["
										+ url + "]");
							}
							Scheme scheme = Scheme.valueOf(portalUrl.getPath());
							actionUrl = createPortletUrl(scheme, portalUrl, escapedUrl);
						} catch (IllegalArgumentException e) {
							LOGGER.severe("Invalid URL type for portlet: protocol :"
									+ url);
							actionUrl = url;
						}
					} else {
						String directLink = portalUrl
								.getParameter(Bridge.DIRECT_LINK);
						if (null != directLink
								&& !Boolean.parseBoolean(directLink)) {
							portalUrl.removeParameter(Bridge.DIRECT_LINK);
						}/*
						 * if the inputURL is an absolute path external to this
						 * portlet application[6.5] return the inputURL
						 * unchanged.
						 */
						if (LOGGER.isLoggable(Level.FINE)) {
							LOGGER.fine("Url out of context, keep unchanged:"
									+ portalUrl.toString());
						}
						actionUrl = escapeUrl(escapedUrl, portalUrl.toString());
					}
					/*
					 * If the inputURL contains the parameter
					 * javax.portlet.faces.DirectLink (with a value of "true")
					 * return an absolute path derived from the inputURL. Don't
					 * remove the DirectLink parameter if it exists[6.6]. If the
					 * inputURL contains the parameter
					 * javax.portlet.faces.DirectLink and its value is false
					 * then remove the javax.portlet.faces.DirectLink parameter
					 * and its value from the query string and continue
					 * processing (using the next step concerning determining
					 * the target of the URL)[6.7].
					 */
				} else {
					String directLink = portalUrl
							.getParameter(Bridge.DIRECT_LINK);
					if (null != directLink) {
						if (Boolean.parseBoolean(directLink)) {
							// make absolute url
							PortletRequest request = getRequest();
							portalUrl.setProtocol(request.getScheme()+":");
							portalUrl.setHost("//"+request.getServerName());
							portalUrl.setPort(request.getServerPort());
							String directUrl = portalUrl.toString();
							return escapeUrl(escapedUrl, directUrl);
						} else {
							portalUrl.removeParameter(Bridge.DIRECT_LINK);
						}
					}
					if (!isInContext(portalUrl)) {
						/*
						 * if the inputURL is an absolute path external to this
						 * portlet application[6.5] return the inputURL
						 * unchanged.
						 */
						if (LOGGER.isLoggable(Level.FINE)) {
							LOGGER.fine("Url out of context, encode by portal:"
									+ portalUrl.toString());
						}
						actionUrl = escapeUrl(escapedUrl, portalUrl.toString());
					} else {
						String pathInContext = calculatePathInContext(portalUrl);
						if (isFacesPath(pathInContext)) {
							portalUrl.setParameter(
									Bridge.FACES_VIEW_ID_PARAMETER,
									getViewIdFromPath(pathInContext));
							actionUrl = createActionUrl(portalUrl, escapedUrl);
						} else {
                            //TODO cleanup before 2.1.1
							portalUrl.setParameter(
									Bridge.NONFACES_TARGET_PATH_PARAMETER, pathInContext);
                            pathInContext = calculatePathInContext(portalUrl);
                            portalUrl.setParameter(
									Bridge.NONFACES_TARGET_PATH_PARAMETER, pathInContext);
							actionUrl = createRenderUrl(portalUrl, escapedUrl);
						}
					}
				}
				actionParameters = portalUrl.getParameters();
			} catch (MalformedURLException e) {
				LOGGER.severe("Malformed action URL :" + url);
				actionUrl = url;
				actionParameters = Collections.emptyMap();
			}
		}
		if (LOGGER.isLoggable(Level.FINE)) {
			LOGGER.fine("Action URL encoded as:" + actionUrl);
		}
		// Store url parameters to reuse in redirect()
		encodedActionUrlParameters.put(actionUrl, actionParameters);
		return actionUrl;
	}

	protected String escapeUrl(boolean escapedUrl, String directUrl) {
		if(escapedUrl){
			directUrl = directUrl.replaceAll("\\&", "&amp;");
		}
		return directUrl;
	}

	private boolean isStrictEscaped(String url) {
		int queryStart = url.indexOf('?');
		if(queryStart < 0){
			return false;
		} else {
			return url.indexOf("&amp;", queryStart)>0;
		}
	}

	protected boolean isFacesPath(String pathInContext) {
		if (null != getServletMappingPrefix()) {
			return pathInContext.startsWith(getServletMappingPrefix());
		} else if (null != getServletMappingSuffix()) {
			return pathInContext.endsWith(getServletMappingSuffix());
		}
		// No Faces preffix/suffix defined, all request came to JSF
		return true;
	}

	protected String createPortletUrl(Scheme protocol, PortalActionURL portalUrl, boolean escape) {
		/*
		 * # The scheme is followed by either the keyword action, render or
		 * resource. render indicates a portlet renderURL should be
		 * encoded[6.8]. action indicates a portlet actionURL should be
		 * encoded[6.9]. resource indicates a portlet resourceURL should be
		 * encoded[6.102]. # Following this url type indicator is an optional
		 * query string. Parameter value pairs in the query string are the
		 * parameters that are to be encoded into the portletURL.
		 * 
		 * To generate a link to a Faces view, encode the view as the value of
		 * either the _jsfBridgeViewId or _jsfBridgeViewPath parameter
		 * (depending on whether you are encoding the viewId or the viewPath).
		 * Targets of such references are run in new empty scopes. An exception
		 * is made when the target is the current view and either of the above
		 * parameters is included in the query string with a value of
		 * _jsfBridgeCurrentView. In this case the url is encoded with the
		 * current render parameters and hence retains access to its
		 * state/scope. In all cases the bridge removes the above parameter(s)
		 * from the query string before generating the encoded url.
		 * 
		 * For a resource url, a Faces view is only encoded if one of the
		 * _jsfBridgeViewId or _jsfBridgeViewPath parameters is included in the
		 * query string, otherwise a nonFaces resource url is generated. The
		 * _jsfBridgeCurrentView value is used as a shortcut to indicate the
		 * resource targets the current view.
		 */
		processJsfViewParameter(portalUrl, Bridge.FACES_VIEW_ID_PARAMETER);
		processJsfViewParameter(portalUrl, Bridge.FACES_VIEW_PATH_PARAMETER);
		switch (protocol) {
		case action:
			return createActionUrl(portalUrl, escape);
		case resource:
			return createResourceUrl(portalUrl, escape);
		case render:
			return createRenderUrl(portalUrl, escape);
		default:
			return portalUrl.toString();
		}
	}

	private void processJsfViewParameter(PortalActionURL portalUrl,
			String facesViewParameter) {
		if (portalUrl.hasParameter(facesViewParameter)) {
			if (Bridge.FACES_USE_CURRENT_VIEW_PARAMETER.equals(portalUrl
					.getParameter(Bridge.FACES_VIEW_ID_PARAMETER))) {
				portalUrl.removeParameter(Bridge.FACES_VIEW_ID_PARAMETER);
				putRenderParametersToUrl(portalUrl);
			}
		}
	}

	protected abstract void putRenderParametersToUrl(PortalActionURL portalUrl);

	protected abstract String createActionUrl(PortalActionURL url, boolean escape);

	protected abstract String createResourceUrl(PortalActionURL portalUrl, boolean escape);

	protected abstract String createRenderUrl(PortalActionURL portalUrl, boolean escape);

	private String unescapeUrl(String actionUrl) {
		// TODO - unescape query string only
		return actionUrl.replaceAll("\\&amp\\;", "&");
	}

	private boolean isSpecialUrl(String actionUrl) {
		return actionUrl.startsWith("#")
				|| actionUrl
						.startsWith(PortletExternalContextImpl.WSRP_REWRITE);
	}

	public String encodeResourceURL(String url) {
		if (null == url) {
			throw new NullPointerException();
		}
		if (LOGGER.isLoggable(Level.FINE)) {
			LOGGER.fine("Encode resource URL:" + url);
		}
		if (!isSpecialUrl(url)) {
			try {
				boolean escapedUrl = isStrictEscaped(url);
				PortalActionURL portalUrl = new PortalActionURL(escapedUrl?unescapeUrl(url):url);
				encodeBackLink(portalUrl);
				// JSR-301 chapter 6.1.3.1 requirements:
				// 1) opaque URL
				if ("portlet:".equals(portalUrl.getProtocol())) {
					String actionURL = encodeResourceUrlAsAction(portalUrl, escapedUrl);
					return actionURL;
				} else if (portalUrl.isOpaque()) {
					if (LOGGER.isLoggable(Level.FINE)) {
						LOGGER.fine("Opaque URL:" + url);
					}
					portalUrl.removeParameter(Bridge.VIEW_LINK);
					return escapeUrl(escapedUrl, portalUrl.toString());
				} else if (!isInContext(portalUrl)) {
					// 2) hierarchial url outside context.
					if (LOGGER.isLoggable(Level.FINE)) {
						LOGGER.fine("Resource URL outside of context:"
								+ portalUrl);
					}
					portalUrl.removeParameter(Bridge.VIEW_LINK);
					return encodeURL(portalUrl, escapedUrl);
				} else if ("true".equalsIgnoreCase(portalUrl
						.getParameter(Bridge.VIEW_LINK))) {
					// 3) hierarchical and targets a resource that is within
					// this application
					portalUrl.removeParameter(Bridge.VIEW_LINK);
					String actionURL = encodeResourceUrlAsAction(portalUrl, escapedUrl);
					return actionURL;
				} else {
					/*
					 * determine whether the targeted resources will be
					 * satisfied using Portlet 2.0 in protocol resource serving
					 * support or Portlet 1.0 out of band http support. The
					 * bridge must encode the resource to be satisfied by the in
					 * protocol resource serving support if:
					 * 
					 * the query string contains the parameter
					 * javax.portlet.faces.InProtocolResourceLink the target is
					 * a Faces resource, one that in a servlet environment would
					 * be processed running through the FacesServlet.
					 * 
					 * 
					 * Note: The bridge can choose the strategy for acquiring
					 * (representing) all other resources though it is noted
					 * that out of band (http) resource access is generally more
					 * efficient.
					 */
					portalUrl.removeParameter(Bridge.VIEW_LINK);
					String inContextPath = calculatePathInContext(portalUrl);
					if(portalUrl.hasParameter(Bridge.IN_PROTOCOL_RESOURCE_LINK)){
						portalUrl.setPath(inContextPath);
						url = createResourceUrl(portalUrl, escapedUrl);
					} else if (isFacesPath(inContextPath)) {
						portalUrl.setParameter(
								Bridge.FACES_VIEW_ID_PARAMETER,
								getViewIdFromPath(inContextPath));
						portalUrl.setPath(inContextPath);
						url = createResourceUrl(portalUrl, escapedUrl);
					} 
					else {
						portalUrl.setPath(getRequestContextPath()+inContextPath);
						url = encodeURL(portalUrl, escapedUrl);
					}
				}
			} catch (MalformedURLException e) {
				LOGGER.severe("Malformed action URL :" + url);
			}
		}
		if (LOGGER.isLoggable(Level.FINE)) {
			LOGGER.fine("Resource URL encoded as:" + url);
		}
		return url;
	}

	protected String calculatePathInContext(PortalActionURL portalURL) {
		String inContextPath;
        String path = portalURL.getPath();
		if (path.startsWith("/")) {
			// absolute path, remove context path from ID.
            //TCK compliance - return the full nonfaces view prepended with context path
            if(portalURL.getParameter(Bridge.NONFACES_TARGET_PATH_PARAMETER) != null){
                return path;
            }
			inContextPath = path.substring(getRequestContextPath().length());
            //return path;
		} else {
			// resolve relative URL aganist current view.
			FacesContext facesContext = FacesContext.getCurrentInstance();
			UIViewRoot viewRoot = facesContext.getViewRoot();
			if (null != viewRoot && null != viewRoot.getViewId()
					&& viewRoot.getViewId().length() > 0) {
				String viewId = viewRoot.getViewId();
				int indexOfSlash = viewId.lastIndexOf('/');
				if (indexOfSlash >= 0) {
					inContextPath = viewId.substring(0, indexOfSlash + 1)
							+ path;
				} else {
					inContextPath = '/' + path;
				}
			} else {
				// No clue where we are
				inContextPath = '/' + path;
			}
		}
		inContextPath = URI.create(inContextPath).normalize().getPath();
		return inContextPath;
	}

	private String encodeResourceUrlAsAction(PortalActionURL portalUrl, boolean escape) {
		// 1. view link. TODO - would it better to create renderURL
		// ?
		portalUrl.removeParameter(Bridge.VIEW_LINK);
		String actionURL = encodeActionURL(escapeUrl(escape, portalUrl.toString()));
		if (LOGGER.isLoggable(Level.FINE)) {
			LOGGER.fine("Resource URL to faces view, encode as actionURL:"
					+ actionURL);
		}
		return actionURL;
	}

	protected boolean isInContext(PortalActionURL portalUrl) {
		String directLink = portalUrl.getParameter(Bridge.DIRECT_LINK);
		if (null != directLink) {
			if (Boolean.parseBoolean(directLink)) {
				return false;
			} else {
				portalUrl.removeParameter(Bridge.DIRECT_LINK);
			}
		}
		return portalUrl.isInContext(getRequestContextPath());
	}

	protected void encodeBackLink(PortalActionURL portalUrl) {
		String backLink = portalUrl.getParameter(Bridge.BACK_LINK);
		if (null != backLink) {
			portalUrl.removeParameter(Bridge.BACK_LINK);
			FacesContext facesContext = FacesContext.getCurrentInstance();
			String viewId;
			if (null != facesContext.getViewRoot()
					&& null != (viewId = facesContext.getViewRoot().getViewId())) {
				ViewHandler viewHandler = facesContext.getApplication()
						.getViewHandler();
				String actionURL = viewHandler.getActionURL(facesContext,
						viewId);
				portalUrl.addParameter(backLink, encodeActionURL(actionURL));
			}
		}
	}

	protected String encodePortletUrl(BaseURL portletURL, boolean escape) {
		StringWriter out = new StringWriter();
	      try {
			portletURL.write(out, escape);
	          return out.toString();
		} catch (IOException e) {
			throw new FacesException(e);
		}
	}

	protected void restoreParametersFromScope() {
		BridgeRequestScope requestScope = this.portletBridgeContext
				.getRequestScope();
		Map<String, String[]> savedRequestParameters = requestScope.getRequestParameters();
		if (savedRequestParameters != null) {
			getExtraRequestParameterValuesMap().putAll(savedRequestParameters);
		}
		String conversationId = requestScope.getConversationId();
		if(null != conversationId){
			getExtraRequestParameterValuesMap().put(requestScope.getConversationIdParameter(), new String[]{conversationId});
		}
	}


}
